# © 2026 The Radiativity Company
# Licensed under the Apache License, Version 2.0
# See the LICENSE file for details.

"""Pharmaceutical compounding example demonstrating advanced Pydantic patterns."""
